from __future__ import annotations

from typing import TYPE_CHECKING

from starlite import ASGIConnection, BaseRouteHandler, NotAuthorizedException

__all__ = ["roles_accepted", "roles_required"]


if TYPE_CHECKING:
    from starlite.types import Guard


def roles_accepted(*roles: str) -> "Guard":
    """Get a [Guard][starlite.types.Guard] callable and inject authorized role names.

    Args:
        roles: Iterable of authorized role names.

    Returns:
        Starlite [Guard][starlite.types.callable_types.Guard] callable
    """

    def roles_accepted_guard(connection: ASGIConnection, _: BaseRouteHandler) -> None:
        """Authorize a request if any of the user's roles matches any of the supplied roles."""

        if any(role.name in roles for role in connection.user.roles):
            return
        raise NotAuthorizedException()

    return roles_accepted_guard


def roles_required(*roles: str) -> "Guard":
    """Get a [Guard][starlite.types.Guard] callable and inject authorized role names.

    Args:
        roles: Iterable of authorized role names.
    """

    def roles_required_guard(connection: ASGIConnection, _: BaseRouteHandler) -> None:
        """Authorize a request if the user's roles matches all of the supplied roles."""
        user_role_names = [role.name for role in connection.user.roles]
        if all(role in user_role_names for role in roles):
            return
        raise NotAuthorizedException()

    return roles_required_guard



from typing import TYPE_CHECKING, Any, Callable, Dict, List, Literal, Optional, Type, Union
from uuid import UUID

from starlite import (
    HTTPRouteHandler,
    Provide,
    Request,
    Response,
    Router,
    delete,
    get,
    patch,
    post,
    put,
)
from starlite.contrib.jwt import JWTAuth, JWTCookieAuth
from starlite.exceptions import ImproperlyConfiguredException, NotAuthorizedException, PermissionDeniedException
from starlite.security.session_auth.auth import SessionAuth

from starlite_users.adapter.sqlalchemy.mixins import UserModelType
from starlite_users.schema import (
    ForgotPasswordSchema,
    ResetPasswordSchema,
    RoleCreateDTOType,
    RoleReadDTOType,
    RoleUpdateDTOType,
    UserAuthSchema,
    UserCreateDTOType,
    UserReadDTOType,
    UserRoleSchema,
    UserUpdateDTOType,
)
from starlite_users.service import BaseUserService

__all__ = [
    "get_auth_handler",
    "get_current_user_handler",
    "get_password_reset_handler",
    "get_registration_handler",
    "get_role_management_handler",
    "get_user_management_handler",
    "get_verification_handler",
]


if TYPE_CHECKING:
    from starlite.types import Guard

IDENTIFIER_URI = "/{id_:uuid}"  # TODO: define via config


def get_registration_handler(
    path: str,
    user_create_dto: Type[UserCreateDTOType],
    user_read_dto: Type[UserReadDTOType],
    service_dependency: Callable,
    tags: Optional[List[str]] = None,
) -> HTTPRouteHandler:
    """Get registration route handlers.

    Args:
        path: The path for the router.
        user_create_dto: A subclass of [BaseUserCreateDTO][starlite_users.schema.BaseUserCreateDTO]
        user_read_dto: A subclass of [BaseUserReadDTO][starlite_users.schema.BaseUserReadDTO]
        service_dependency: Callable to provide a `UserService` instance.
        tags: A list of string tags to append to the schema of the route handler.
    """

    @post(path, dependencies={"service": Provide(service_dependency)}, exclude_from_auth=True, tags=tags)
    async def register(data: user_create_dto, service: BaseUserService) -> user_read_dto:  # type: ignore[valid-type]
        """Register a new user."""

        user = await service.register(data)
        return user_read_dto.from_orm(user)

    return register


def get_verification_handler(
    path: str, user_read_dto: Type[UserReadDTOType], service_dependency: Callable, tags: Optional[List[str]] = None
) -> HTTPRouteHandler:
    """Get verification route handlers.

    Args:
        path: The path for the router.
        user_read_dto: A subclass of [BaseUserReadDTO][starlite_users.schema.BaseUserReadDTO]
        service_dependency: Callable to provide a `UserService` instance.
        tags: A list of string tags to append to the schema of the route handler.
    """

    @post(path, dependencies={"service": Provide(service_dependency)}, exclude_from_auth=True, tags=tags)
    async def verify(token: str, service: BaseUserService) -> user_read_dto:  # type: ignore[valid-type]
        """Verify a user with a given JWT."""

        user = await service.verify(token)
        return user_read_dto.from_orm(user)

    return verify


def get_auth_handler(
    login_path: str,
    logout_path: str,
    user_read_dto: Type[UserReadDTOType],
    service_dependency: Callable,
    auth_backend: Union[JWTAuth, JWTCookieAuth, SessionAuth],
    tags: Optional[List[str]] = None,
) -> Router:
    """Get authentication/login route handlers.

    Args:
        login_path: The path for the login router.
        logout_path: The path for the logout router.
        user_read_dto: A subclass of [BaseUserReadDTO][starlite_users.schema.BaseUserReadDTO]
        service_dependency: Callable to provide a `UserService` instance.
        auth_backend: A Starlite authentication backend.
        tags: A list of string tags to append to the schema of the route handlers.
    """

    @post(login_path, dependencies={"service": Provide(service_dependency)}, exclude_from_auth=True, tags=tags)
    async def login_session(data: UserAuthSchema, service: BaseUserService, request: Request) -> user_read_dto:  # type: ignore[valid-type]
        """Authenticate a user."""
        if not isinstance(auth_backend, SessionAuth):
            raise ImproperlyConfiguredException("session login can only be used with SesssionAuth")

        user = await service.authenticate(data)
        if user is None:
            request.clear_session()
            raise NotAuthorizedException(detail="login failed, invalid input")

        if user.is_verified is False:
            raise PermissionDeniedException(detail="not verified")

        request.set_session({"user_id": user.id})
        return user_read_dto.from_orm(user)

    @post(login_path, dependencies={"service": Provide(service_dependency)}, exclude_from_auth=True, tags=tags)
    async def login_jwt(data: UserAuthSchema, service: BaseUserService) -> Response[user_read_dto]:  # type: ignore
        """Authenticate a user."""

        if not isinstance(auth_backend, (JWTAuth, JWTCookieAuth)):
            raise ImproperlyConfiguredException("jwt login can only be used with JWTAuth")

        user = await service.authenticate(data)
        if user is None:
            raise NotAuthorizedException(detail="login failed, invalid input")

        if user.is_verified is False:
            raise PermissionDeniedException(detail="not verified")

        user_dto = user_read_dto.from_orm(user)
        return auth_backend.login(identifier=str(user.id), response_body=user_dto)

    @post(logout_path, tags=tags)
    async def logout(request: Request) -> None:
        """Log an authenticated user out."""
        request.clear_session()

    route_handlers = []
    if isinstance(auth_backend, SessionAuth):
        route_handlers.extend([login_session, logout])
    else:
        route_handlers.append(login_jwt)

    return Router(path="/", route_handlers=route_handlers)  # type: ignore[arg-type]


def get_current_user_handler(
    path: str,
    user_read_dto: Type[UserReadDTOType],
    user_update_dto: Type[UserUpdateDTOType],
    service_dependency: Callable,
    tags: Optional[List[str]] = None,
) -> Router:
    """Get current-user route handlers.

    Args:
        path: The path for the router.
        user_read_dto: A subclass of [BaseUserReadDTO][starlite_users.schema.BaseUserReadDTO]
        user_update_dto: A subclass of [BaseUserUpdateDTO][starlite_users.schema.BaseUserUpdateDTO]
        service_dependency: Callable to provide a `UserService` instance.
        tags: A list of string tags to append to the schema of the route handlers.
    """

    @get(path, tags=tags)
    async def get_current_user(request: Request[UserModelType, Dict[Literal["user_id"], str]]) -> user_read_dto:  # type: ignore[valid-type]
        """Get current user info."""

        return user_read_dto.from_orm(request.user)

    @put(path, dependencies={"service": Provide(service_dependency)}, tags=tags)
    async def update_current_user(
        data: user_update_dto,  # type: ignore[valid-type]
        request: Request[UserModelType, Dict[Literal["user_id"], str]],
        service: BaseUserService,
    ) -> user_read_dto:  # type: ignore[valid-type]
        """Update the current user."""

        updated_user = await service.update_user(id_=request.user.id, data=data)
        return user_read_dto.from_orm(updated_user)

    return Router(path="/", route_handlers=[get_current_user, update_current_user])


def get_password_reset_handler(
    forgot_path: str, reset_path: str, service_dependency: Callable, tags: Optional[List[str]] = None
) -> Router:
    """Get forgot-password and reset-password route handlers.

    Args:
        forgot_path: The path for the forgot-password router.
        reset_path: The path for the reset-password router.
        service_dependency: Callable to provide a `UserService` instance.
        tags: A list of string tags to append to the schema of the route handlers.
    """

    @post(forgot_path, dependencies={"service": Provide(service_dependency)}, exclude_from_auth=True, tags=tags)
    async def forgot_password(data: ForgotPasswordSchema, service: BaseUserService) -> None:
        await service.initiate_password_reset(data.email)
        return

    @post(reset_path, dependencies={"service": Provide(service_dependency)}, exclude_from_auth=True, tags=tags)
    async def reset_password(data: ResetPasswordSchema, service: BaseUserService) -> None:
        await service.reset_password(data.token, data.password)
        return

    return Router(path="/", route_handlers=[forgot_password, reset_password])


def get_user_management_handler(
    path_prefix: str,
    guards: List["Guard"],
    opt: Dict[str, Any],
    user_read_dto: Type[UserReadDTOType],
    user_update_dto: Type[UserUpdateDTOType],
    service_dependency: Callable,
    tags: Optional[List[str]] = None,
) -> Router:
    """Get user management route handlers.

    Note:
        Routes are guarded by role authorization.

    Args:
        path_prefix: The path prefix for the routers.
        guards: List of Guard callables to determine who is authorized to manage users.
        opt: Optional route handler 'opts' to provide additional context to Guards.
        user_read_dto: A subclass of [BaseUserReadDTO][starlite_users.schema.BaseUserReadDTO]
        user_update_dto: A subclass of [BaseUserUpdateDTO][starlite_users.schema.BaseUserUpdateDTO]
        service_dependency: Callable to provide a `UserService` instance.
        tags: A list of string tags to append to the schema of the route handlers.
    """

    @get(IDENTIFIER_URI, guards=guards, opt=opt, dependencies={"service": Provide(service_dependency)}, tags=tags)
    async def get_user(id_: UUID, service: BaseUserService) -> user_read_dto:  # type: ignore[valid-type]
        """Get a user by id."""

        user = await service.get_user(id_)
        return user_read_dto.from_orm(user)

    @put(IDENTIFIER_URI, guards=guards, opt=opt, dependencies={"service": Provide(service_dependency)}, tags=tags)
    async def update_user(
        id_: UUID, data: user_update_dto, service: BaseUserService  # type: ignore[valid-type]
    ) -> user_read_dto:  # type: ignore[valid-type]
        """Update a user's attributes."""

        user = await service.update_user(id_, data)
        return user_read_dto.from_orm(user)

    @delete(IDENTIFIER_URI, guards=guards, opt=opt, dependencies={"service": Provide(service_dependency)}, tags=tags)
    async def delete_user(id_: UUID, service: BaseUserService) -> None:
        """Delete a user from the database."""

        return await service.delete_user(id_)

    return Router(path=path_prefix, route_handlers=[get_user, update_user, delete_user])


def get_role_management_handler(
    path_prefix: str,
    assign_role_path: str,
    revoke_role_path: str,
    guards: List["Guard"],
    opt: Dict[str, Any],
    role_create_dto: Type[RoleCreateDTOType],
    role_read_dto: Type[RoleReadDTOType],
    role_update_dto: Type[RoleUpdateDTOType],
    user_read_dto: Type[UserReadDTOType],
    service_dependency: Callable,
    tags: Optional[List[str]] = None,
) -> Router:
    """Get role management route handlers.

    Note:
        Routes are guarded by role authorization.

    Args:
        path_prefix: The path prefix for the routers.
        assign_role_path: The path for the role assignment router.
        revoke_role_path: The path for the role revokement router.
        guards: List of Guard callables to determine who is authorized to manage roles.
        opt: Optional route handler 'opts' to provide additional context to Guards.
        role_create_dto: A subclass of [BaseRoleCreateDTO][starlite_users.schema.BaseRoleCreateDTO]
        role_read_dto: A subclass of [BaseRoleReadDTO][starlite_users.schema.BaseRoleReadDTO]
        role_update_dto: A subclass of [BaseRoleUpdateDTO][starlite_users.schema.BaseRoleUpdateDTO]
        user_read_dto: A subclass of [BaseUserReadDTO][starlite_users.schema.BaseUserReadDTO]
        service_dependency: Callable to provide a `UserService` instance.
        tags: A list of string tags to append to the schema of the route handlers.
    """

    @post(guards=guards, opt=opt, dependencies={"service": Provide(service_dependency)}, tags=tags)
    async def create_role(data: role_create_dto, service: BaseUserService) -> role_read_dto:  # type: ignore[valid-type]
        """Create a new role."""
        role = await service.add_role(data)
        return role_read_dto.from_orm(role)

    @put(IDENTIFIER_URI, guards=guards, opt=opt, dependencies={"service": Provide(service_dependency)}, tags=tags)
    async def update_role(id_: UUID, data: role_update_dto, service: BaseUserService) -> role_read_dto:  # type: ignore[valid-type]
        """Update a role in the database."""

        role = await service.update_role(id_, data)
        return role_read_dto.from_orm(role)

    @delete(IDENTIFIER_URI, guards=guards, opt=opt, dependencies={"service": Provide(service_dependency)}, tags=tags)
    async def delete_role(id_: UUID, service: BaseUserService) -> None:
        """Delete a role from the database."""

        return await service.delete_role(id_)

    @patch(
        path=assign_role_path, guards=guards, opt=opt, dependencies={"service": Provide(service_dependency)}, tags=tags
    )
    async def assign_role_to_user(data: UserRoleSchema, service: BaseUserService) -> user_read_dto:  # type: ignore[valid-type]
        """Assign a role to a user."""

        user = await service.assign_role_to_user(data.user_id, data.role_id)
        return user_read_dto.from_orm(user)

    @patch(
        path=revoke_role_path, guards=guards, opt=opt, dependencies={"service": Provide(service_dependency)}, tags=tags
    )
    async def revoke_role_from_user(data: UserRoleSchema, service: BaseUserService) -> user_read_dto:  # type: ignore[valid-type]
        """Revoke a role from a user."""

        user = await service.revoke_role_from_user(data.user_id, data.role_id)
        return user_read_dto.from_orm(user)

    return Router(
        path_prefix, route_handlers=[create_role, assign_role_to_user, revoke_role_from_user, update_role, delete_role]
    )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Generic, Literal

from pydantic import SecretStr
from starlite.contrib.jwt import JWTAuth, JWTCookieAuth
from starlite.exceptions import ImproperlyConfiguredException
from starlite.security.session_auth import SessionAuth

from starlite_users.adapter.sqlalchemy.mixins import (
    SQLAlchemyRoleMixin,
    UserModelType,
)
from starlite_users.adapter.sqlalchemy.repository import SQLAlchemyUserRepository
from starlite_users.schema import (
    BaseRoleCreateDTO,
    BaseRoleReadDTO,
    BaseRoleUpdateDTO,
    BaseUserCreateDTO,
    BaseUserReadDTO,
    BaseUserUpdateDTO,
)
from starlite_users.user_handlers import (
    get_jwt_retrieve_user_handler,
    get_session_retrieve_user_handler,
)

__all__ = [
    "AuthHandlerConfig",
    "CurrentUserHandlerConfig",
    "PasswordResetHandlerConfig",
    "RegisterHandlerConfig",
    "RoleManagementHandlerConfig",
    "StarliteUsersConfig",
    "UserManagementHandlerConfig",
    "VerificationHandlerConfig",
]

if TYPE_CHECKING:
    from starlite.middleware.session.base import BaseBackendConfig
    from starlite.types import Guard

    from starlite_users.service import BaseUserService


@dataclass
class AuthHandlerConfig:
    """Configuration for user authentication route handlers.

    Passing an instance to `StarliteUsersConfig` will automatically take care of handler registration on the app.
    """

    login_path: str = "/login"
    """The path for the user authentication/login route."""
    logout_path: str = "/logout"
    """The path for the logout route."""
    tags: list[str] | None = None
    """A list of string tags to append to the schema of the route handler(s)."""


@dataclass
class CurrentUserHandlerConfig:
    """Configuration for the current-user route handler.

    Passing an instance to `StarliteUsersConfig` will automatically take care of handler registration on the app.
    """

    path: str = "/users/me"
    """The path to get or update the currently logged-in user."""
    tags: list[str] | None = None
    """A list of string tags to append to the schema of the route handler(s)."""


@dataclass
class PasswordResetHandlerConfig:
    """Configuration for the forgot-password and reset-password route handlers.

    Passing an instance to `StarliteUsersConfig` will automatically take care of handler registration on the app.
    """

    forgot_path: str = "/forgot-password"
    """The path for the forgot-password route."""
    reset_path: str = "/reset-password"
    """The path for the reset-password route."""
    tags: list[str] | None = None
    """A list of string tags to append to the schema of the route handler(s)."""


@dataclass
class RegisterHandlerConfig:
    """Configuration for the user registration route handler.

    Passing an instance to `StarliteUsersConfig` will automatically take care of handler registration on the app.
    """

    path: str = "/register"
    """The path for the registration/signup route."""
    tags: list[str] | None = None
    """A list of string tags to append to the schema of the route handler(s)."""


@dataclass
class RoleManagementHandlerConfig:
    """Configuration for the role management route handlers.

    Passing an instance to `StarliteUsersConfig` will automatically take care of handler registration on the app.
    """

    path_prefix: str = "/users/roles"
    """The prefix for the router path."""
    assign_role_path: str = "/assign"
    """The path for the role assignment router."""
    revoke_role_path: str = "/revoke"
    """The path for the role revokement router."""
    guards: list[Guard] = field(default_factory=list)
    """A list of callable [Guards][starlite.types.Guard] that determines who is authorized to manage roles."""
    opt: dict[str, Any] = field(default_factory=dict)
    """Optional route handler [opts][starlite.controller.Controller.opt] to provide additional context to Guards."""
    tags: list[str] | None = None
    """A list of string tags to append to the schema of the route handler(s)."""


@dataclass
class UserManagementHandlerConfig:
    """Configuration for user management (read, update, delete) route handlers.

    Passing an instance to `StarliteUsersConfig` will automatically take care of handler registration on the app.

    Note:
    - These routes make use of Starlite `Guard`s to require authorization. Callers require admin or similar privileges.
    """

    path_prefix: str = "/users"
    """The prefix for the router path.

    By default, the path will be suffixed with `'/{id_:uuid}'`.
    """
    guards: list[Guard] = field(default_factory=list)
    """A list of callable [Guards][starlite.types.Guard] that determines who is authorized to manage other users."""
    opt: dict[str, Any] = field(default_factory=dict)
    """Optional route handler [opts][starlite.controller.Controller.opt] to provide additional context to Guards."""
    tags: list[str] | None = None
    """A list of string tags to append to the schema of the route handler(s)."""


@dataclass
class VerificationHandlerConfig:
    """Configuration for the user verification route handler.

    Passing an instance to `StarliteUsersConfig` will automatically take care of handler registration on the app.
    """

    path: str = "/verify"
    """The path for the verification route."""
    tags: list[str] | None = None
    """A list of string tags to append to the schema of the route handler(s)."""


@dataclass
class StarliteUsersConfig(Generic[UserModelType]):
    """Configuration class for StarliteUsers."""

    auth_backend: Literal["session", "jwt", "jwt_cookie"]
    """The authentication backend to use by Starlite."""
    secret: SecretStr
    """Secret string for securely signing tokens."""
    user_model: type[UserModelType]
    """A subclass of a `User` ORM model."""
    user_service_class: type[BaseUserService]
    """A subclass of [BaseUserService][starlite_users.service.BaseUserService]."""
    user_create_dto: type[BaseUserCreateDTO] = BaseUserCreateDTO
    """A subclass of [BaseUserCreateDTO][starlite_users.schema.BaseUserCreateDTO]."""
    user_read_dto: type[BaseUserReadDTO] = BaseUserReadDTO
    """A subclass of [BaseUserReadDTO][starlite_users.schema.BaseUserReadDTO]."""
    user_update_dto: type[BaseUserUpdateDTO] = BaseUserUpdateDTO
    """A subclass of [BaseUserUpdateDTO][starlite_users.schema.BaseUserUpdateDTO]."""
    user_repository_class: type[SQLAlchemyUserRepository] = SQLAlchemyUserRepository
    """The user repository class to use."""
    auth_exclude_paths: list[str] = field(default_factory=lambda: ["/schema"])
    """Paths to be excluded from authentication checks."""
    hash_schemes: list[str] = field(default_factory=lambda: ["argon2"])
    """Schemes to use for password encryption.

    Defaults to `["argon2"]`
    """
    session_backend_config: BaseBackendConfig | None = None
    """Optional backend configuration for session based authentication.

    Notes:
        - Required if `auth_backend` is set to `session`.
    """
    role_model: type[SQLAlchemyRoleMixin] = SQLAlchemyRoleMixin
    """A subclass of a `Role` ORM model.

    Notes:
        - Required if `role_management_handler_config` is set.
    """
    role_create_dto: type[BaseRoleCreateDTO] = BaseRoleCreateDTO
    """A subclass of [BaseRoleCreateDTO][starlite_users.schema.BaseRoleCreateDTO].

    Notes:
        - Required if `role_management_handler_config` is set.
    """
    role_read_dto: type[BaseRoleReadDTO] = BaseRoleReadDTO
    """A subclass of [BaseRoleReadDTO][starlite_users.schema.BaseRoleReadDTO].

    Notes:
        - Required if `role_management_handler_config` is set.
    """
    role_update_dto: type[BaseRoleUpdateDTO] = BaseRoleUpdateDTO
    """A subclass of [BaseRoleUpdateDTO][starlite_users.schema.BaseRoleUpdateDTO].

    Notes:
        - Required if `role_management_handler_config` is set.
    """
    auth_handler_config: AuthHandlerConfig | None = None
    """Optional instance of [AuthHandlerConfig][starlite_users.config.AuthHandlerConfig]. If set, registers the route
    handler(s) on the app.

    Note:
        - At least one route handler config must be set.
    """
    current_user_handler_config: CurrentUserHandlerConfig | None = None
    """Optional current-user route handler configuration. If set, registers the route handler(s) on the app.

    Note:
        - At least one route handler config must be set.
    """
    password_reset_handler_config: PasswordResetHandlerConfig | None = None
    """Optional password reset route handler configuration. If set, registers the route handler(s) on the app.

    Note:
        - At least one route handler config must be set.
    """
    register_handler_config: RegisterHandlerConfig | None = None
    """Optional registration/signup route handler configuration. If set, registers the route handler(s) on the app.

    Note:
        - At least one route handler config must be set.
    """
    role_management_handler_config: RoleManagementHandlerConfig | None = None
    """Optional role management route handler configuration. If set, registers the route handler(s) on the app.

    Note:
        - At least one route handler config must be set.
    """
    user_management_handler_config: UserManagementHandlerConfig | None = None
    """Optional user management route handler configuration. If set, registers the route handler(s) on the app.

    Note:
        - At least one route handler config must be set.
    """
    verification_handler_config: VerificationHandlerConfig | None = None
    """Optional user verification route handler configuration. If set, registers the route handler(s) on the app.

    Note:
        - At least one route handler config must be set.
    """
    _auth_config: JWTAuth | JWTCookieAuth | SessionAuth | None = None

    def __post_init__(self) -> None:
        """Validate the configuration.

        - A session backend must be configured if `auth_backend` is set to `'session'`.
        - At least one route handler must be configured.
        - `role_model`, `role_create_dto`, `role_read_dto` and `role_update_dto` are required fields if `role_management_handler_config` is configured.
        """
        if self.auth_backend == "session" and not self.session_backend_config:
            raise ImproperlyConfiguredException(
                'session_backend_config must be set when auth_backend is set to "session"'
            )
        handler_configs = [
            "auth_handler_config",
            "current_user_handler_config",
            "password_reset_handler_config",
            "register_handler_config",
            "role_management_handler_config",
            "user_management_handler_config",
            "verification_handler_config",
        ]
        if isinstance(self.secret, str):
            self.secret = SecretStr(self.secret)
        if len(self.secret) not in [16, 24, 32]:
            raise ImproperlyConfiguredException("secret must be 16, 24 or 32 characters")
        if all(getattr(self, config) is None for config in handler_configs):
            raise ImproperlyConfiguredException("at least one route handler must be configured")
        if self.role_management_handler_config and self.role_model is SQLAlchemyRoleMixin:
            raise ImproperlyConfiguredException("role_model must be set when role_management_handler_config is set")

        self._auth_config = self._get_auth_config()

    @property
    def auth_config(self) -> JWTAuth | JWTCookieAuth | SessionAuth:
        return self._auth_config or self._get_auth_config()

    def _get_auth_config(self) -> JWTAuth | JWTCookieAuth | SessionAuth:
        if self.auth_backend == "session":
            return SessionAuth(  # pyright: ignore
                retrieve_user_handler=get_session_retrieve_user_handler(
                    user_model=self.user_model,
                    role_model=self.role_model,
                    user_repository_class=self.user_repository_class,
                ),
                session_backend_config=self.session_backend_config,  # type: ignore[arg-type]
                exclude=self.auth_exclude_paths,
            )
        if self.auth_backend == "jwt":
            return JWTAuth(  # pyright: ignore
                retrieve_user_handler=get_jwt_retrieve_user_handler(
                    user_model=self.user_model,
                    role_model=self.role_model,
                    user_repository_class=self.user_repository_class,
                ),
                token_secret=self.secret.get_secret_value(),
                exclude=self.auth_exclude_paths,
            )

        return JWTCookieAuth(  # pyright: ignore
            retrieve_user_handler=get_jwt_retrieve_user_handler(
                user_model=self.user_model,
                role_model=self.role_model,
                user_repository_class=self.user_repository_class,
            ),
            token_secret=self.secret.get_secret_value(),
            exclude=self.auth_exclude_paths,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast

from sqlalchemy import select
from sqlalchemy.exc import IntegrityError, NoResultFound  # type: ignore[attr-defined]
from starlite.exceptions import ImproperlyConfiguredException

from starlite_users.adapter.sqlalchemy.mixins import (
    RoleModelType,
    SQLAlchemyRoleMixin,
    UserModelType,
)
from starlite_users.exceptions import (
    RepositoryConflictException,
    RepositoryNotFoundException,
)
from starlite_users.generics import AbstractUserRepository

__all__ = ["SQLAlchemyUserRepository"]


if TYPE_CHECKING:
    from uuid import UUID

    from sqlalchemy.ext.asyncio import AsyncSession


class SQLAlchemyUserRepository(
    AbstractUserRepository[UserModelType, RoleModelType]
):  # TODO: create generic base for piccolo, tortoise etc
    """SQLAlchemy implementation of user persistence layer."""

    user_model: type[UserModelType]

    def __init__(
        self,
        session: "AsyncSession",
        user_model: type[UserModelType],
        role_model: type[RoleModelType],
    ) -> None:
        """Initialise a repository instance.

        Args:
            session: A SQLAlchemy `AsyncSession`.
            user_model: A subclass of [SQLAlchemyUserMixin][starlite_users.adapter.sqlalchemy.mixins.SQLAlchemyUserMixin]
            role_model: A subclass of [SQLAlchemyRoleMixin][starlite_users.adapter.sqlalchemy.mixins.SQLAlchemyRoleMixin]
        """
        self.session = session
        self.user_model = user_model
        self.role_model = role_model

    async def add_user(self, user: UserModelType) -> UserModelType:
        """Add a user to the database.

        Args:
            user: A SQLAlchemy User model instance.

        Raises:
            RepositoryConflictException: when the given user already exists.
        """
        try:
            self.session.add(user)
            await self.session.flush()
            await self.session.refresh(user)

            await self.session.commit()

            return user
        except IntegrityError as e:
            raise RepositoryConflictException from e

    async def get_user(self, id_: "UUID") -> UserModelType:
        """Retrieve a user from the database by id.

        Args:
            id_: UUID corresponding to a user primary key.

        Raises:
            RepositoryNotFoundException: when no user matches the query.
        """
        result = await self.session.execute(select(self.user_model).where(self.user_model.id == id_))  # type: ignore[arg-type]
        try:
            return cast("UserModelType", result.unique().scalar_one())
        except NoResultFound as e:
            raise RepositoryNotFoundException from e

    async def get_user_by(self, **kwargs: Any) -> UserModelType:
        """Retrieve a user from the database by arbitrary keyword arguments.

        Args:
            **kwargs: Keyword arguments to pass as filters.

        Examples:
            ```python
            repository = SQLAlchemyUserRepository(...)
            john = await repository.get_user_by(email="john@example.com")
            ```

        Raises:
            RepositoryNotFoundException: when no user matches the query.
        """
        result = await self.session.execute(
            select(self.user_model).where(*(getattr(self.user_model, k) == v for k, v in kwargs.items()))  # type: ignore[arg-type]
        )
        try:
            return cast("UserModelType", result.unique().scalar_one())
        except NoResultFound as e:
            raise RepositoryNotFoundException from e

    async def update_user(self, id_: "UUID", data: dict[str, Any]) -> UserModelType:
        """Update arbitrary user attributes in the database.

        Args:
            id_: UUID corresponding to a user primary key.
            data: Dictionary to map to user columns and values.
        """
        user = await self.get_user(id_)
        return await self._update(user, data)

    async def delete_user(self, id_: "UUID") -> None:
        """Delete a user from the database.

        Args:
            id_: UUID corresponding to a user primary key.
        """
        user = await self.get_user(id_)
        await self.session.delete(user)
        await self.session.commit()

    async def _update(self, user: UserModelType, data: dict[str, Any]) -> UserModelType:
        for attr, val in data.items():
            setattr(user, attr, val)

        await self.session.commit()
        return user

    async def add_role(self, role: RoleModelType) -> RoleModelType:
        """Add a role to the database.

        Args:
            role: A SQLAlchemy Role model instance.

        Raises:
            RepositoryConflictException: when the given role already exists.
        """
        try:
            self.session.add(role)
            await self.session.flush()
            await self.session.refresh(role)

            await self.session.commit()

            return role
        except IntegrityError as e:
            raise RepositoryConflictException from e

    async def assign_role_to_user(self, user: UserModelType, role: RoleModelType) -> UserModelType:
        """Add a role to a user.

        Args:
            user: The user to receive the role.
            role: The role to add to the user.
        """
        user.roles.append(role)
        await self.session.commit()
        return user

    async def revoke_role_from_user(self, user: UserModelType, role: RoleModelType) -> UserModelType:
        """Revoke a role to a user.

        Args:
            user: The user to revoke the role from.
            role: The role to revoke from the user.
        """
        user.roles.remove(role)
        await self.session.commit()
        return user

    async def get_role(self, id_: "UUID") -> RoleModelType:
        """Retrieve a role from the database by id.

        Args:
            id_: UUID corresponding to a role primary key.

        Raises:
            RepositoryNotFoundException: when no role matches the query.
        """
        if self.role_model is None or not issubclass(self.role_model, SQLAlchemyRoleMixin):
            raise ImproperlyConfiguredException("StarliteUsersConfig.role_model must subclass SQLAlchemyRoleMixin")
        result = await self.session.execute(select(self.role_model).where(self.role_model.id == id_))  # type: ignore[arg-type]
        try:
            return cast("RoleModelType", result.unique().scalar_one())
        except NoResultFound as e:
            raise RepositoryNotFoundException from e

    async def get_role_by_name(self, name: str) -> RoleModelType:
        """Retrieve a role from the database by name.

        Args:
            name: The name of the desired role record.

        Raises:
            RepositoryNotFoundException: when no role matches the query.
        """
        if self.role_model is None or not issubclass(self.role_model, SQLAlchemyRoleMixin):
            raise ImproperlyConfiguredException("StarliteUsersConfig.role_model must subclass SQLAlchemyRoleMixin")
        result = await self.session.execute(select(self.role_model).where(self.role_model.name == name))  # type: ignore[arg-type]
        try:
            return cast("RoleModelType", result.unique().scalar_one())
        except NoResultFound as e:
            raise RepositoryNotFoundException from e

    async def update_role(self, id_: "UUID", data: dict[str, Any]) -> RoleModelType:
        """Update arbitrary role attributes in the database.

        Args:
            id_: UUID corresponding to a role primary key.
            data: Dictionary to map to role columns and values.
        """
        role = await self.get_role(id_)
        for attr, val in data.items():
            setattr(role, attr, val)

        await self.session.commit()
        return role

    async def delete_role(self, id_: "UUID") -> None:
        """Delete a role from the database.

        Args:
            id_: UUID corresponding to a role primary key.
        """
        role = await self.get_role(id_)
        await self.session.delete(role)
        await self.session.commit()



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Sequence

from sqlalchemy.orm import Session
from starlite import State  # noqa: TCH002
from starlite.exceptions import ImproperlyConfiguredException
from starlite.plugins.sql_alchemy import SQLAlchemyPlugin
from starlite.types import Scope  # noqa: TCH002

__all__ = ["get_service_dependency"]


if TYPE_CHECKING:
    from pydantic import SecretStr

    from starlite_users.adapter.sqlalchemy.mixins import RoleModelType, UserModelType
    from starlite_users.adapter.sqlalchemy.repository import SQLAlchemyUserRepository
    from starlite_users.service import UserServiceType


def get_service_dependency(
    user_model: type["UserModelType"],
    user_service_class: type["UserServiceType"],
    user_repository_class: type[SQLAlchemyUserRepository],
    role_model: type[RoleModelType],
    secret: SecretStr,
    hash_schemes: Sequence[str] | None,
) -> Callable:
    """Get a service dependency callable.

    Args:
        user_model: A subclass of a `User` ORM model.
        role_model: A subclass of a `Role` ORM model.
        user_service_class: A subclass of [BaseUserService][starlite_users.service.BaseUserService]
        user_repository_class: A subclass of `BaseUserRepository` to use for database operations.
        secret: Secret string for securely signing tokens.
        hash_schemes: Schemes to use for password encryption.
    """

    def get_service(scope: Scope, state: State) -> UserServiceType:
        """Instantiate service and repository for use with DI.

        Args:
            scope: ASGI scope
            state: The application.state instance
        """
        session = None
        if not any(isinstance(plugin, SQLAlchemyPlugin) for plugin in scope["app"].plugins):
            raise ImproperlyConfiguredException("SQLAlchemyPlugin must be configured with SQLAlchemyConfig")
        for plugin in scope["app"].plugins:
            if isinstance(plugin, SQLAlchemyPlugin):
                if plugin._config is None:
                    raise ImproperlyConfiguredException("SQLAlchemyPlugin must be configured with SQLAlchemyConfig")
                session = plugin._config.create_db_session_dependency(state, scope)
                break

        if session is None or isinstance(session, Session):
            raise ImproperlyConfiguredException("session must be instance of sqlalchemy.AsyncSession")

        repository = user_repository_class(session=session, user_model=user_model, role_model=role_model)

        return user_service_class(repository=repository, secret=secret, hash_schemes=hash_schemes)

    return get_service



from typing import List, Optional, TypeVar
from uuid import UUID

from pydantic import BaseModel, SecretStr

__all__ = [
    "BaseRoleCreateDTO",
    "BaseRoleReadDTO",
    "BaseRoleUpdateDTO",
    "BaseUserCreateDTO",
    "BaseUserReadDTO",
    "BaseUserRoleReadDTO",
    "BaseUserUpdateDTO",
    "ForgotPasswordSchema",
    "ResetPasswordSchema",
    "UserAuthSchema",
    "UserRoleSchema",
]


class BaseRoleReadDTO(BaseModel):
    """Base role read schema."""

    id: UUID
    name: str
    description: str

    class Config:
        orm_mode = True


class BaseRoleCreateDTO(BaseModel):
    """Base role create schema."""

    name: str
    description: str


class BaseRoleUpdateDTO(BaseModel):
    """Base role update schema."""

    name: Optional[str]
    description: Optional[str]


class BaseUserReadDTO(BaseModel):
    """Base user read schema."""

    id: UUID
    email: str
    is_active: bool
    is_verified: bool

    class Config:
        orm_mode = True


class BaseUserRoleReadDTO(BaseUserReadDTO):
    """Base user read schema that includes `roles`."""

    roles: List[Optional[BaseRoleReadDTO]]


class BaseUserCreateDTO(BaseModel):
    """Base user create schema."""

    email: str
    password: SecretStr
    is_active: Optional[bool]
    is_verified: Optional[bool]


class BaseUserUpdateDTO(BaseModel):
    """Base user update schema."""

    email: Optional[str]
    password: Optional[SecretStr]
    is_active: Optional[bool]
    is_verified: Optional[bool]


class UserAuthSchema(BaseModel):
    """User authentication schema."""

    email: str
    password: SecretStr


class ForgotPasswordSchema(BaseModel):
    """Forgot password schema."""

    email: str


class ResetPasswordSchema(BaseModel):
    """Reset password schema."""

    token: str
    password: SecretStr


class UserRoleSchema(BaseModel):
    """User role association schema."""

    user_id: UUID
    role_id: UUID


RoleCreateDTOType = TypeVar("RoleCreateDTOType", bound=BaseRoleCreateDTO)
RoleReadDTOType = TypeVar("RoleReadDTOType", bound=BaseRoleReadDTO)
RoleUpdateDTOType = TypeVar("RoleUpdateDTOType", bound=BaseRoleUpdateDTO)

UserCreateDTOType = TypeVar("UserCreateDTOType", bound=BaseUserCreateDTO)
UserReadDTOType = TypeVar("UserReadDTOType", bound=BaseUserReadDTO)
UserUpdateDTOType = TypeVar("UserUpdateDTOType", bound=BaseUserUpdateDTO)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Sequence

from starlite_users.dependencies import get_service_dependency
from starlite_users.exceptions import (
    RepositoryException,
    TokenException,
    repository_exception_handler,
    token_exception_handler,
)
from starlite_users.route_handlers import (
    get_auth_handler,
    get_current_user_handler,
    get_password_reset_handler,
    get_registration_handler,
    get_role_management_handler,
    get_user_management_handler,
    get_verification_handler,
)

__all__ = ["StarliteUsers"]


if TYPE_CHECKING:
    from starlite import HTTPRouteHandler, Request, Response, Router
    from starlite.config import AppConfig
    from starlite.contrib.jwt import JWTAuth, JWTCookieAuth
    from starlite.security.session_auth import SessionAuth

    from starlite_users.config import StarliteUsersConfig


class StarliteUsers:
    """A Starlite extension for authentication, authorization and user management."""

    def __init__(self, config: "StarliteUsersConfig") -> None:
        """Construct a StarliteUsers instance."""
        self._config = config

    def on_app_init(self, app_config: "AppConfig") -> "AppConfig":
        """Register routers, auth strategies etc on the Starlite app.

        Args:
            app_config: An instance of [AppConfig][starlite.config.AppConfig]
        """
        auth_backend = self._config.auth_config
        route_handlers = self._get_route_handlers(auth_backend)

        app_config = auth_backend.on_app_init(app_config)
        app_config.route_handlers.extend(route_handlers)

        exception_handlers: dict[type[Exception], Callable[[Request, Any], Response]] = {
            TokenException: token_exception_handler,
            RepositoryException: repository_exception_handler,
        }
        app_config.exception_handlers.update(exception_handlers)  # type: ignore[arg-type]

        return app_config

    def _get_route_handlers(
        self, auth_backend: JWTAuth | JWTCookieAuth | SessionAuth
    ) -> Sequence[HTTPRouteHandler | Router]:
        """Parse the route handler configs to get Routers."""

        handlers: list[HTTPRouteHandler | Router] = []
        service_dependency_provider = get_service_dependency(
            user_model=self._config.user_model,
            role_model=self._config.role_model,
            user_service_class=self._config.user_service_class,
            user_repository_class=self._config.user_repository_class,
            secret=self._config.secret,
            hash_schemes=self._config.hash_schemes,
        )
        if self._config.auth_handler_config:
            handlers.append(
                get_auth_handler(
                    login_path=self._config.auth_handler_config.login_path,
                    logout_path=self._config.auth_handler_config.logout_path,
                    user_read_dto=self._config.user_read_dto,
                    service_dependency=service_dependency_provider,
                    auth_backend=auth_backend,
                    tags=self._config.auth_handler_config.tags,
                )
            )
        if self._config.current_user_handler_config:
            handlers.append(
                get_current_user_handler(
                    path=self._config.current_user_handler_config.path,
                    user_read_dto=self._config.user_read_dto,
                    user_update_dto=self._config.user_update_dto,
                    service_dependency=service_dependency_provider,
                    tags=self._config.current_user_handler_config.tags,
                )
            )
        if self._config.password_reset_handler_config:
            handlers.append(
                get_password_reset_handler(
                    forgot_path=self._config.password_reset_handler_config.forgot_path,
                    reset_path=self._config.password_reset_handler_config.reset_path,
                    service_dependency=service_dependency_provider,
                    tags=self._config.password_reset_handler_config.tags,
                )
            )
        if self._config.register_handler_config:
            handlers.append(
                get_registration_handler(
                    path=self._config.register_handler_config.path,
                    user_create_dto=self._config.user_create_dto,
                    user_read_dto=self._config.user_read_dto,
                    service_dependency=service_dependency_provider,
                    tags=self._config.register_handler_config.tags,
                )
            )
        if self._config.role_management_handler_config:
            handlers.append(
                get_role_management_handler(
                    path_prefix=self._config.role_management_handler_config.path_prefix,
                    assign_role_path=self._config.role_management_handler_config.assign_role_path,
                    revoke_role_path=self._config.role_management_handler_config.revoke_role_path,
                    guards=self._config.role_management_handler_config.guards,
                    opt=self._config.role_management_handler_config.opt,
                    role_create_dto=self._config.role_create_dto,
                    role_read_dto=self._config.role_read_dto,
                    role_update_dto=self._config.role_update_dto,
                    user_read_dto=self._config.user_read_dto,
                    service_dependency=service_dependency_provider,
                    tags=self._config.role_management_handler_config.tags,
                )
            )
        if self._config.user_management_handler_config:
            handlers.append(
                get_user_management_handler(
                    path_prefix=self._config.user_management_handler_config.path_prefix,
                    guards=self._config.user_management_handler_config.guards,
                    opt=self._config.user_management_handler_config.opt,
                    user_read_dto=self._config.user_read_dto,
                    user_update_dto=self._config.user_update_dto,
                    service_dependency=service_dependency_provider,
                    tags=self._config.user_management_handler_config.tags,
                )
            )
        if self._config.verification_handler_config:
            handlers.append(
                get_verification_handler(
                    path=self._config.verification_handler_config.path,
                    user_read_dto=self._config.user_read_dto,
                    service_dependency=service_dependency_provider,
                    tags=self._config.verification_handler_config.tags,
                )
            )
        return handlers



from __future__ import annotations

from typing import TypeVar
from uuid import UUID, uuid4

from sqlalchemy import Boolean, Column, String
from sqlalchemy.orm.attributes import Mapped  # type: ignore[attr-defined]  # noqa: TCH002
from sqlalchemy.orm.decl_api import declarative_mixin

from starlite_users.adapter.sqlalchemy.guid import GUID

__all__ = ["SQLAlchemyRoleMixin", "SQLAlchemyUserMixin"]


@declarative_mixin
class SQLAlchemyUserMixin:
    """Base SQLAlchemy user mixin."""

    id: Mapped[UUID] = Column(
        GUID(),
        primary_key=True,
        default=uuid4,
        nullable=False,
    )
    email: Mapped[str] = Column(String(320), nullable=False, unique=True)
    password_hash: Mapped[str] = Column(String(1024))
    is_active: Mapped[bool] = Column(Boolean(), nullable=False, default=False)
    is_verified: Mapped[bool] = Column(Boolean(), nullable=False, default=False)

    @property
    def roles(self) -> Mapped[list["SQLAlchemyRoleMixin"]]:
        """Dummy placeholder."""
        return []


@declarative_mixin
class SQLAlchemyRoleMixin:
    """Base SQLAlchemy role mixin."""

    id: Mapped[UUID] = Column(
        GUID(),
        primary_key=True,
        default=uuid4,
        nullable=False,
    )
    name: Mapped[str] = Column(String(255), nullable=False, unique=True)
    description: Mapped[str] = Column(String(255), nullable=True)


UserModelType = TypeVar("UserModelType", bound=SQLAlchemyUserMixin)
RoleModelType = TypeVar("RoleModelType", bound=SQLAlchemyRoleMixin)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable

from starlite.exceptions import ImproperlyConfiguredException
from starlite.plugins.sql_alchemy.plugin import SQLAlchemyPlugin

from starlite_users.exceptions import RepositoryNotFoundException

__all__ = ["get_jwt_retrieve_user_handler", "get_session_retrieve_user_handler"]


if TYPE_CHECKING:
    from starlite import ASGIConnection
    from starlite.contrib.jwt import Token

    from starlite_users.adapter.sqlalchemy.mixins import RoleModelType, UserModelType
    from starlite_users.generics import AbstractUserRepository


def get_session_retrieve_user_handler(
    user_model: type[UserModelType],
    role_model: type[RoleModelType],
    user_repository_class: type[AbstractUserRepository],
) -> Callable:
    """Get retrieve_user_handler functions for session backends.

    Args:
        user_model: A subclass of a `User` ORM model.
        role_model: A subclass of a `Role` ORM model.
        user_repository_class: A subclass of `BaseUserRepository` to use for database operations.
    """

    async def retrieve_user_handler(
        session: dict[str, Any], connection: ASGIConnection[Any, Any, Any]
    ) -> UserModelType | None:
        """Get a user from the database based on session info.

        Args:
            session: Starlite session.
            connection: The ASGI connection.
        """
        db_config = None
        for plugin in connection.app.plugins:
            if isinstance(plugin, SQLAlchemyPlugin):
                db_config = plugin._config
                break
        if db_config is None:
            raise ImproperlyConfiguredException("sqlalchemy plugin is not registered")

        async_session = db_config.create_db_session_dependency(state=connection.app.state, scope=connection.scope)

        repository = user_repository_class(session=async_session, user_model=user_model, role_model=role_model)
        try:
            user = await repository.get_user(session.get("user_id", ""))
            if user.is_active and user.is_verified:
                return user  # type: ignore[no-any-return]
        except RepositoryNotFoundException:
            return None
        return None

    return retrieve_user_handler


def get_jwt_retrieve_user_handler(
    user_model: type[UserModelType],
    role_model: type[RoleModelType],
    user_repository_class: type[AbstractUserRepository],
) -> Callable:
    """Get retrieve_user_handler functions for jwt backends.

    Args:
        user_model: A subclass of a `User` ORM model.
        role_model: A subclass of a `Role` ORM model.
        user_repository_class: A subclass of `BaseUserRepository` to use for database operations.
    """

    async def retrieve_user_handler(token: Token, connection: ASGIConnection[Any, Any, Any]) -> user_model | None:  # type: ignore[valid-type]
        """Get a user from the database based on JWT info.

        Args:
            token: Encoded JWT.
            connection: The ASGI connection.
        """
        db_config = None
        for plugin in connection.app.plugins:
            if isinstance(plugin, SQLAlchemyPlugin):
                db_config = plugin._config
                break
        if db_config is None:
            raise ImproperlyConfiguredException("sqlalchemy plugin is not registered")

        async_session = db_config.create_db_session_dependency(state=connection.app.state, scope=connection.scope)

        repository = user_repository_class(session=async_session, user_model=user_model, role_model=role_model)
        try:
            user = await repository.get_user(token.sub)
            if user.is_active and user.is_verified:
                return user  # type: ignore[no-any-return]
        except RepositoryNotFoundException:
            return None
        return None

    return retrieve_user_handler



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, Protocol, TypeVar

if TYPE_CHECKING:
    from uuid import UUID

__all__ = ["AbstractUserRepository", "RoleProtocol", "UserProtocol"]


class UserProtocol(Protocol):
    """Base protocol to be inherited when implementing user model mixins."""

    id: UUID
    email: str
    password_hash: str
    is_active: bool
    is_verified: bool


class RoleProtocol(Protocol):
    """Base protocol to be inherited when implementing role model mixins."""

    id: UUID
    name: str
    description: str


UserT = TypeVar("UserT", bound=UserProtocol)
RoleT = TypeVar("RoleT", bound=RoleProtocol)


class AbstractUserRepository(ABC, Generic[UserT, RoleT]):
    """Base protocol to be inherited when implementing repositories."""

    user_model: type[UserT]
    role_model: type[RoleT]

    def __init__(self, **kwargs: Any) -> None:
        ...

    @abstractmethod
    async def add_user(self, user: UserT) -> UserT:
        """Add a user to the database.

        Args:
            user: A SQLAlchemy User model instance.

        Raises:
            RepositoryConflictException: when the given user already exists.
        """
        raise NotImplementedError()

    @abstractmethod
    async def get_user(self, id_: UUID) -> UserT:
        """Retrieve a user from the database by id.

        Args:
            id_: UUID corresponding to a user primary key.

        Raises:
            RepositoryNotFoundException: when no user matches the query.
        """
        raise NotImplementedError()

    @abstractmethod
    async def get_user_by(self, **kwargs: Any) -> UserT:
        """Retrieve a user from the database by arbitrary keyword arguments.

        Args:
            **kwargs: Keyword arguments to pass as filters.

        Examples:
            ```python
            repository = SQLAlchemyUserRepository(...)
            john = await repository.get_user_by(email="john@example.com")
            ```

        Raises:
            RepositoryNotFoundException: when no user matches the query.
        """
        raise NotImplementedError()

    @abstractmethod
    async def update_user(self, id_: UUID, data: dict[str, Any]) -> UserT:
        """Update arbitrary user attributes in the database.

        Args:
            id_: UUID corresponding to a user primary key.
            data: Dictionary to map to user columns and values.
        """
        raise NotImplementedError()

    @abstractmethod
    async def delete_user(self, id_: UUID) -> None:
        """Delete a user from the database.

        Args:
            id_: UUID corresponding to a user primary key.
        """
        raise NotImplementedError()

    @abstractmethod
    async def _update(self, user: UserT, data: dict[str, Any]) -> UserT:
        raise NotImplementedError()

    @abstractmethod
    async def add_role(self, role: RoleT) -> RoleT:
        """Add a role to the database.

        Args:
            role: A SQLAlchemy Role model instance.

        Raises:
            RepositoryConflictException: when the given role already exists.
        """
        raise NotImplementedError()

    @abstractmethod
    async def assign_role_to_user(self, user: UserT, role: RoleT) -> UserT:
        """Add a role to a user.

        Args:
            user: The user to receive the role.
            role: The role to add to the user.
        """
        raise NotImplementedError()

    @abstractmethod
    async def revoke_role_from_user(self, user: UserT, role: RoleT) -> UserT:
        """Revoke a role to a user.

        Args:
            user: The user to revoke the role from.
            role: The role to revoke from the user.
        """
        raise NotImplementedError()

    @abstractmethod
    async def get_role(self, id_: UUID) -> RoleT:
        """Retrieve a role from the database by id.

        Args:
            id_: UUID corresponding to a role primary key.

        Raises:
            RepositoryNotFoundException: when no role matches the query.
        """
        raise NotImplementedError()

    @abstractmethod
    async def get_role_by_name(self, name: str) -> RoleT:
        """Retrieve a role from the database by name.

        Args:
            name: The name of the desired role record.

        Raises:
            RepositoryNotFoundException: when no role matches the query.
        """
        raise NotImplementedError()

    @abstractmethod
    async def update_role(self, id_: UUID, data: dict[str, Any]) -> RoleT:
        """Update arbitrary role attributes in the database.

        Args:
            id_: UUID corresponding to a role primary key.
            data: Dictionary to map to role columns and values.
        """
        raise NotImplementedError()

    @abstractmethod
    async def delete_role(self, id_: UUID) -> None:
        """Delete a role from the database.

        Args:
            id_: UUID corresponding to a role primary key.
        """
        raise NotImplementedError()
